/*
 * Copyright (c) 2013 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-ipv6-0.c,v 1.17 2014/04/19 15:49:46 ca Exp $")

#include "sm/debug.h"
#include "sm/heap.h"
#include "sm/rpool.h"
#include "sm/test.h"
#include "sm/net.h"
#include "sm/str.h"
#include "sm/io.h"

static int Verbose;

#if MTA_NETINET6
struct t6_S {
	const char *t6_in;
	const char *t6_out;
};
typedef struct t6_S t6_T;

#if MTA_INET6_COMPRESSED
/* compressed */
static t6_T ipv6s[] = {
	{ "1234:5678:9abc:def0:fedc:dead:f00f:0101", "1234:5678:9abc:def0:fedc:dead:f00f:101" },
	{ "1080:0:0:0:8:800:200C:417A",	"1080::8:800:200c:417a" },
	{ "FF01:0:0:0:0:0:0:43",	"ff01::43" },
	{ "0:0:0:0:0:0:0:1", "::1" },
	{ "0:0:0:0:0:0:0:0", "::" },
	{ "[1080:0:0:0:8:800:200C:417A]",	"1080::8:800:200c:417a" },
	{ "[FF01:0:0:0:0:0:0:43]",	"ff01::43" },
	{ "[0:0:0:0:0:0:0:1]", "::1" },
	{ "[::1]", "::1" },
	{ "[0:0:0:0:0:0:0:0]", "::" },
	{ NULL, NULL }
};
#endif

/* uncompressed */
static t6_T ipv6s2[] = {
	{ "1080:0:0:0:8:800:200C:417A ",	"1080:0:0:0:8:800:200c:417a" },
	{ "FF01:0:0:0:0:0:0:43,",	"ff01:0:0:0:0:0:0:43" },
	{ "0:0:0:0:0:0:0:1;", "0:0:0:0:0:0:0:1" },
	{ "0:0:0:0:0:0:0:0\n", "0:0:0:0:0:0:0:0" },
	{ "[1080:0:0:0:8:800:200C:417A],",	"1080:0:0:0:8:800:200c:417a" },
	{ "[FF01:0:0:0:0:0:0:43];",	"ff01:0:0:0:0:0:0:43" },
	{ "[0:0:0:0:0:0:0:1]\n", "0:0:0:0:0:0:0:1" },
	{ "[0:0:0:0:0:0:0:0] ", "0:0:0:0:0:0:0:0" },
	{ NULL, NULL }
};

#if MTA_NETINET6 >= 2
/* uncompressed */
static t6_T ipv6s4[] = {
	{ "0:0:0:0:0:FFFF:8190:3426 ", "129.144.52.38" },
	{ NULL, NULL }
};
#endif


/* "full" */
static char *ipv6f[] = {
	"1234:5678:9abc:def0:fedc:dead:f00f:101",
	"1080:0:0:0:8:800:200c:417a",
	"ff01:0:0:0:0:0:0:43",
	"0:0:0:0:0:0:0:1",
	"0:0:0:0:0:0:0:0",
	NULL
};

static void
test(void)
{
	ipv6_T ipv6;
	t6_T *ipv6e;
	sm_ret_T ret;
	int i;
	char *ip, *endptr, ipv6str[INET6_ADDRSTRLEN];
	sm_str_P str, rvrs;
	sm_sockaddr_T sockaddr;
	sm_ipaddr_T ipaddr;

	str = rvrs = NULL;
	str = sm_str_new(NULL, 64, 100);
	SM_TEST_E(str != NULL);
	rvrs = sm_str_new(NULL, 100, 128);
	SM_TEST_E(rvrs != NULL);

#if MTA_INET6_COMPRESSED
	for (ipv6e = ipv6s; ipv6e->t6_in != NULL; ipv6e++) {
		(void) strlcpy(ipv6str, ipv6e->t6_in, sizeof(ipv6str));
		ret = sm_inet_a2ipv6(ipv6str, &endptr, &ipv6);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(endptr == ipv6str + strlen(ipv6e->t6_in));
		sm_str_clr(str);
		ret = sm_inet_ipv6str(&ipv6, str);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(strcmp((char *) sm_str_getdata(str), ipv6e->t6_out) == 0);
		if (Verbose > 1)
			sm_io_fprintf(smioerr, "in=%s, out=%S, exp=%s\n",
				ipv6str, str, ipv6e->t6_out);
		sm_snprintf(ipv6str, sizeof(ipv6str), "%B", &ipv6);
		SM_TEST(strcmp(ipv6str, ipv6e->t6_out) == 0);

	}
#endif

	for (ipv6e = ipv6s2; ipv6e->t6_in != NULL; ipv6e++) {
		(void) strlcpy(ipv6str, ipv6e->t6_in, sizeof(ipv6str));
		ret = sm_inet_a2ipv6(ipv6str, &endptr, &ipv6);
		SM_TEST(ret == SM_SUCCESS);
		if (Verbose > 1 && ret != SM_SUCCESS)
			sm_io_fprintf(smioerr, "in=\"%s\", ret=%x\n", ipv6str, ret);
		SM_TEST(endptr == ipv6str + strlen(ipv6e->t6_in) - 1);
		sm_str_clr(str);
		ret = sm_inet_ipv6fstr(&ipv6, str);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(strcmp((char *) sm_str_getdata(str), ipv6e->t6_out) == 0);
		if (Verbose > 1)
			sm_io_fprintf(smioerr, "in=\"%s\", out=\"%S\", exp=\"%s\", end=\"%s\", ok=%d\n",
				ipv6str, str, ipv6e->t6_out, endptr,
				strcmp((char *) sm_str_getdata(str), ipv6e->t6_out) == 0);

#if !MTA_INET6_COMPRESSED
		sm_snprintf(ipv6str, sizeof(ipv6str), "%B", &ipv6);
		SM_TEST(strcmp(ipv6str, ipv6e->t6_out) == 0);
		if (Verbose > 1 && strcmp(ipv6str, ipv6e->t6_out) != 0)
			sm_io_fprintf(smioerr, "out=\"%s\", exp=\"%s\"\n",
				ipv6str, ipv6e->t6_out);
#endif

		sm_str_clr(str);
		sockaddr.sa.sa_family = AF_INET6;
		IPV6_ASSIGN(&sockaddr.sin6.sin6_addr, &ipv6);
		ret = sm_sockaddr2str(&sockaddr, str);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(strcmp((char *) sm_str_getdata(str), ipv6e->t6_out) == 0);

		(void) strlcpy(ipv6str, ipv6e->t6_in, sizeof(ipv6str));
		ret = sm_inet_a2sockaddr(ipv6str, &endptr, &sockaddr);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(endptr == ipv6str + strlen(ipv6e->t6_in) - 1);
		SM_TEST(sockaddr.sa.sa_family == AF_INET6);
		sm_str_clr(str);
		ret = sm_sockaddr2str(&sockaddr, str);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(strcmp((char *) sm_str_getdata(str), ipv6e->t6_out) == 0);
		if (Verbose > 1)
			sm_io_fprintf(smioerr, "in=%s, out=%S, exp=%s\n",
				ipv6str, str, ipv6e->t6_out);

		sm_str_clr(str);
		IP_ASSIGN_SCK(ipaddr, sockaddr);
		ret = sm_ipaddr2str(&ipaddr, str);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(strcmp((char *) sm_str_getdata(str), ipv6e->t6_out) == 0);
		if (Verbose > 1)
			sm_io_fprintf(smioerr, "in=%s, out=%S, exp=%s\n",
				ipv6str, str, ipv6e->t6_out);

	}

#if MTA_NETINET6 >= 2
	for (ipv6e = ipv6s4; ipv6e->t6_in != NULL; ipv6e++) {
		(void) strlcpy(ipv6str, ipv6e->t6_in, sizeof(ipv6str));
		ret = sm_inet_a2ipv6(ipv6str, &endptr, &ipv6);
		SM_TEST(ret == SM_SUCCESS);
		if (Verbose > 1 && ret != SM_SUCCESS)
			sm_io_fprintf(smioerr, "in=\"%s\", ret=%x\n", ipv6str, ret);
		SM_TEST(endptr == ipv6str + strlen(ipv6e->t6_in) - 1);
		sm_str_clr(str);

		(void) strlcpy(ipv6str, ipv6e->t6_in, sizeof(ipv6str));
		ret = sm_inet_a2sockaddr(ipv6str, &endptr, &sockaddr);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(endptr == ipv6str + strlen(ipv6e->t6_in) - 1);
		SM_TEST(sockaddr.sa.sa_family == AF_INET6);
		sm_str_clr(str);
		ret = sm_sockaddr2str(&sockaddr, str);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(strcmp((char *) sm_str_getdata(str), ipv6e->t6_out) == 0);
		if (Verbose > 1)
			sm_io_fprintf(smioerr, "in=%s, out=%S, exp=%s\n",
				ipv6str, str, ipv6e->t6_out);

		sm_str_clr(str);
		IP_ASSIGN_SCK(ipaddr, sockaddr);
		ret = sm_ipaddr2str(&ipaddr, str);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(strcmp((char *) sm_str_getdata(str), ipv6e->t6_out) == 0);
		if (Verbose > 1)
			sm_io_fprintf(smioerr, "in=%s, out=%S, exp=%s\n",
				ipv6str, str, ipv6e->t6_out);
	}
#endif /* MTA_NETINET6 >= 2 */

	sm_memzero(&ipv6, sizeof(ipv6));
	sm_str_clr(rvrs);
	ret = sm_inet_ipv62arpastr(&ipv6, rvrs);
	SM_TEST_E(ret == SM_SUCCESS);
	SM_TEST(strcmp((char *) sm_str_getdata(rvrs),
		"0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa")
		== 0);

	if (Verbose > 1)
		sm_io_fprintf(smioerr, "%S\n", rvrs);

	for (i = 0; (ip = ipv6f[i]) != NULL; i++) {
		(void) strlcpy(ipv6str, ip, sizeof(ipv6str));
		ret = sm_inet_a2ipv6(ipv6str, &endptr, &ipv6);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(endptr == ipv6str + strlen(ip));
		sm_str_clr(str);
		ret = sm_inet_ipv6fstr(&ipv6, str);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(strcmp((char *) sm_str_getdata(str), ip) == 0);
		if (Verbose > 1 ||
		   (Verbose > 0 && strcmp((char *) sm_str_getdata(str), ip) != 0))
			sm_io_fprintf(smioerr, "out=%S, exp=%s\n", str, ip);
	}

#if 0
	ret = sm_inet_arpa2ipv6((const char *)sm_str_getdata(rvrs), &ipv6_2);
	SM_TEST_E(ret == SM_SUCCESS);
	SM_TEST_E(ipv6 == ipv6_2);

	i = UINT32_MAX;
	sm_str_clr(str);
	ret = sm_inet_ipv6str(i, str);
	SM_TEST(strcmp((char *) sm_str_getdata(str), "255.255.255.255") == 0);
	SM_TEST(ret == SM_SUCCESS);
	ret = sm_inet_a2ipv6((const char *)sm_str_getdata(str), NULL, &ipv6);
	SM_TEST(ret == SM_SUCCESS);
	SM_TEST(ipv6 == i);

	sm_str_clr(rvrs);
	ret = sm_inet_ipv62arpastr(i, rvrs);
	SM_TEST(strcmp((char *) sm_str_getdata(rvrs),
			"255.255.255.255.in-addr.arpa") == 0);
	ret = sm_inet_arpa2ipv6((const char *)sm_str_getdata(rvrs), &ipv6);
	SM_TEST(ret == SM_SUCCESS);
	SM_TEST(ipv6 == i);

	SM_STR_FREE(str);
	SM_STR_FREE(rvrs);

	/* test "overflow" (str too short) */
	str = sm_str_new(NULL, 8, 8);
	SM_TEST_E(str != NULL);
	sm_str_clr(str);
	i = UINT32_MAX;
	ret = sm_inet_ipv6str(i, str);
	SM_TEST(ret != SM_SUCCESS);
	SM_STR_FREE(str);

	for (ip = ipv6f; ip != NULL; ip++) {
		(void) strlcpy(ipv6str, ip, sizeof(ipv6str));
		ret = sm_inet_a2ipv6(ipv6str, &endptr, &ipv6);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(endptr == ipv6str + strlen(ip));
		sm_str_clr(str);
		ret = sm_ipv6str(ip, str);
		SM_TEST(ret == SM_SUCCESS);
		SM_TEST(strcmp((char *) sm_str_getdata(str), ip) == 0);
		if (Verbose > 1)
			sm_io_fprintf(smioerr, "in=%s, out=%S, exp=%s\n",
				ipv6str, str, ip);
	}


	/* test "overflow": str too short but can be expanded */
	str = sm_str_new(NULL, 8, 32);
	SM_TEST_E(str != NULL);
	sm_str_clr(str);
	i = UINT32_MAX;
	ret = sm_inet_ipv6str(i, str);
	SM_TEST(strcmp((char *) sm_str_getdata(str), "255.255.255.255") == 0);
	SM_TEST(ret == SM_SUCCESS);
#endif

  error:
	SM_STR_FREE(str);
	SM_STR_FREE(rvrs);
}
#endif /* MTA_NETINET6 */

int
main(int argc, char *argv[])
{
	int c;

	opterr = 0;
	Verbose = 0;
	while ((c = getopt(argc, argv, "a:V")) != -1) {
		switch (c) {
		  case 'a':
#if MTA_NETINET6
			{
			ipv6_T ipv6;
			sm_ret_T ret;

			if ((ret = sm_inet_a2ipv6(optarg, NULL, &ipv6)) == SM_SUCCESS)
				sm_io_fprintf(smioerr, "%s -> %B\n", optarg, &ipv6);
			else
				sm_io_fprintf(smioerr, "%s: %x\n", optarg, ret);
			exit(0);
			}
#else
			break;
#endif
		  case 'V':
			++Verbose;
			break;
		  default:
			exit(1);
		}
	}
	sm_test_begin(argc, argv, "test ipv6 conversions");
#if MTA_NETINET6
	argc -= optind;
	argv += optind;
	test();
#endif
	return sm_test_end();
}
