/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: sm-conf-type-ip.c,v 1.3 2012/11/05 15:24:57 ca Exp $")

#if SM_LIBCONF_ALONE
#include <limits.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include "sm-conf.h"
#include "sm-util.h"
#else /* SM_LIBCONF_ALONE */
#include "sm/limits.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/memops.h"
#include "sm/string.h"
#include "sm/sm-conf.h"
#include "sm/net.h"
#endif /* SM_LIBCONF_ALONE */

#include "sm-conf-node.h"
#include "sm-conf-state.h"
#include "sm-conf-type.h"
#include "sm-conf-util.h"

/*
**  SM_MEMTOIP --
**
**	Parameters:
**		text -- the text to be converted
**		text_n -- # of bytes pointed to by text
**		endptr -- if non-NULL, a pointer to the first
**			unconverted byte is stored here.
**		val_out -- value is stored here.
**
**	Returns:
**		EINVAL for a format error
**		ERANGE for overflow
**		0 for success.
*/

static sm_ret_T
sm_memtoip(
	char const		*text,
	size_t			text_n,
	char			**endptr,
	sm_ipaddr_P		val_out)
{
	size_t n;
	sm_ret_T ret;
	sm_ipaddr_T ipaddr;
	char ipstr[INET6_ADDRSTRLEN];

	if (text_n >= sizeof(ipstr))
		n = sizeof(ipstr) - 1;
	else
		n = text_n;
	sm_memcpy(ipstr, text, n);
	ipstr[n] = '\0';
	ret = sm_inet_a2ip(ipstr, endptr, &ipaddr);
	memcpy(val_out, &ipaddr, sizeof(sm_ipaddr_T));
	return ret;
}

/* BROKEN XXX */
static int
ip_write(
	sm_conf_definition_T const	*def,
	void			*data,
	sm_ipaddr_P		val)
{
	SM_IS_CONF_DEF(def);
	if (def->scd_size != sizeof(sm_ipaddr_T))
		return SM_CONF_ERR_INVALID;

	memcpy(data, val, sizeof(sm_ipaddr_T));
	return 0;
}

static int
ip_scan(
	sm_conf_T		*smc,
	sm_conf_node_T const	*node,
	sm_conf_definition_T const *def,
	char const		*text,
	size_t			text_n,
	sm_ipaddr_P		val_out)
{
	char		loc[SM_CONF_ERROR_BUFFER_SIZE];
	size_t		i;
	int			err;
	char		*endptr;

	SM_IS_CONF_DEF(def);

	err = sm_memtoip(text, text_n, &endptr, (sm_ipaddr_P) val_out);
	if (err != 0) {
		sm_conf_error_add(smc, "%s: expected IP address, got '%.*s'",
			sm_conf_node_location(smc, node, loc, sizeof loc),
			(int)text_n, text);
		return SM_CONF_ERR_TYPE;
	}

	i = endptr - text;

	/* If there's something left, and we have suffixes, use them. */
	if (i < text_n) {
		sm_conf_error_add(smc, "%s: unexpected trailing "
			"characters in '%.*s'",
			sm_conf_node_location(smc, node, loc, sizeof loc),
			(int)text_n, text);
		return SM_CONF_ERR_TYPE;
	}

	return 0;
}

static int
sm_conf_type_ip_node_to_value(
	sm_conf_T			*smc,
	sm_conf_definition_T const	*def,
	uint32_t			 rqmnts,
	sm_conf_node_T		*node,
	void				*data)
{
	char const		*text;
	size_t			text_n;
	int				err;
	sm_ipaddr_T		val;

	SM_IS_CONF_DEF(def);
	err = sm_conf_node_to_value(smc, "ipaddrean", node, &text, &text_n);
	if (err != 0) return err;

	err = ip_scan(smc, node, def, text, text_n, &val);
	if (err != 0) return err;

	if (data != NULL)
		err = ip_write(def, data, &val);

	return err;
}

static int
sm_conf_type_ip_value_check(
	sm_conf_T			*smc,
	sm_conf_definition_T const	*def,
	void const			*data)
{
	SM_IS_CONF_DEF(def);

	/* if we have a check function, use it. */
	if (def->scd_check != NULL)
		return (* def->scd_check)(smc, def->scd_check_data, def, data);
	return 0;
}

static int
sm_conf_type_ip_value_null(
	sm_conf_T			*smc,
	sm_conf_definition_T const	*def,
	void				*data)
{
	sm_ipaddr_T		ipaddr;
	int				err;

	if (data != NULL) {
		SM_IS_CONF_DEF(def);
		sm_memzero(&ipaddr, sizeof(ipaddr));

		/* is there a default? if NULL: don't set value to 0 */
		if (def->scd_default != NULL) {
			/* is there an actual value? */
			if (*def->scd_default != '\0') {
				err = ip_scan(smc, NULL, def, def->scd_default,
					strlen(def->scd_default), &ipaddr);
				if (err != 0) return err;
			}
			err = ip_write(def, data, &ipaddr);
			if (err != 0)
				return err;
		}
	}
	return 0;
}

sm_conf_type_T const
sm_conf_type_ip_data =
{
	sm_conf_type_ip_node_to_value,
	sm_conf_type_ip_value_check,
	sm_conf_type_ip_value_null
};
