#include "sm/generic.h"
SM_RCSID("@(#)$Id: metaconf.c,v 1.3 2013/12/03 03:20:54 ca Exp $")
#include "sm/common.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#undef MTA_USE_SASL
#include "sm/sysexits.h"
#include "sm/sscnfdef.h"
#include "sm/sm-conf.h"
#include "sm/sm-conf-prt.h"

const char *progname;
const char *conf_name = "/etc/meta1/meta1.conf";

static void
usage(void)
{
	sm_io_fprintf(smioerr, "%s: Extract values from MeTA1 configuration file\n",
	       progname);
	sm_io_fprintf(smioerr, "Usage: %s [-heinsq] [-f FILE] [-S CHAR] VAR [VAR...]\n\n",
	       progname);
	sm_io_fprintf(smioerr, "VAR is a fully qualified name of a configuration variable\n"
		      "e.g. smtps.policy_milter.socket.address\n\n");
	sm_io_fprintf(smioerr, "Options are:\n");
	sm_io_fprintf(smioerr, "  -h        Display this help summary\n"
	       "  -e        Exit with code 1 if any of VARs is not defined\n"
	       "  -i        Display variable indexes\n"
	       "  -n        Display variable names\n"
	       "  -S CHAR   Use CHAR as a separator\n"
	       "  -s        Skip undefined variables\n"
	       "  -q        Quote values on output\n"
	       "  -f FILE   Use FILE instead of %s\n",
	       conf_name);

	exit(0);
}

static void
usage_error(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	sm_io_fprintf(smioerr, "%s: ", progname);
	sm_io_vfprintf(smioerr, fmt, ap);
	sm_io_fprintf(smioerr, "Try %s -h for more information\n", progname);
	va_end(ap);
	exit(EX_USAGE);
}

static void *
emalloc(size_t size)
{
	void *p = malloc(size);
	if (!p) {
		sm_io_fprintf(smioerr, "%m", ENOMEM);
		exit(EX_SOFTWARE);
	}
	return p;
}

static void *
ecalloc(size_t count, size_t size)
{
	void *p = calloc(count, size);
	if (!p) {
		sm_io_fprintf(smioerr, "%m", ENOMEM);
		exit(EX_SOFTWARE);
	}
	return p;
}

/* Return length of the shortest prefix in STR.  Prefixes are delimited
   by dots. */
static int
conf_prefix_len(const char *str)
{
	return strcspn(str, ".");
}

/* Find longest matching prefix in A and B.  If found, return 0 and store
   prefix length in *PLEN.  Otherwise, return lexical difference between
   the two strings (as strcmp(3) does) */
static int
conf_prefix_cmp(const char *a, const char *b, size_t *plen)
{
	int i;
	int diff;
	for (i = 0; (diff = (unsigned char)a[i] - (unsigned char)b[i]) == 0;
	     i++) {
		if (a[i] == '.' || a[i] == 0) {
			if (plen)
				*plen = i;
			return 0;
		}
	}
	return diff;
}

/* Configuration lookup structure. */
struct conf_lookup {
	int idx;            /* Index in the array */
	char *kw;           /* Keyword */
	char *val;          /* Value */
};

/* Compare two struct conf_lookup by their keywords. */
static int
conf_lookup_kw_cmp(const void *a, const void *b)
{
	const char *sa = ((struct conf_lookup *)a)->kw;
	const char *sb = ((struct conf_lookup *)b)->kw;
	int rc;
	size_t preflen;

	while ((rc = conf_prefix_cmp(sa, sb, &preflen)) == 0) {
		sa += preflen;
		sb += preflen;
		if (*sa == 0)
			return 0;
		sa++;
		sb++;
	}
	return rc;
}

/* Compare two struct conf_lookup by their indexes. */
static int
conf_lookup_idx_cmp(const void *a, const void *b)
{
	return ((struct conf_lookup *)a)->idx -
		((struct conf_lookup *)b)->idx;
}

/* Return the number of unique keyword prefixes in first LACNT elements
   of the array LA.  Ignore first BASEPREFLEN characters of each keyword. */
static size_t
count_uniq_prefixes(struct conf_lookup *la, size_t lacnt,
		    size_t basepreflen)
{
	size_t i;
	size_t count = 1;
	char *base_pref = la->kw;
	char *pref = base_pref + basepreflen;
	size_t preflen = conf_prefix_len(pref) + 1;

	for (i = 1, la++; i < lacnt; i++, la++) {
		char *kw = la->kw;
		size_t kwlen = strlen(kw);

		if (kwlen + 1 < basepreflen
		    || memcmp(kw, base_pref, basepreflen))
			break;
		kw += basepreflen;
		kwlen -= basepreflen;
		if (kwlen + 1 < preflen || memcmp(kw, pref, preflen)) {
			count++;
			pref = kw;
			preflen = conf_prefix_len(pref) + 1;
		}
	}
	return count;
}

/* Create a configuration definition for LACNT lookup elements from
   the array LOOKUP.  Ignore first PREFLEN characters of each keyword.
   Use BASE to compute values of scd_offset member. */
static sm_conf_definition_T *
create_def(void *base,
	   struct conf_lookup *lookup, size_t lacnt, size_t preflen)
{
	size_t i, j;
	size_t num;
	sm_conf_definition_T *def;

	if (lacnt == 0)
		return NULL;

	num = count_uniq_prefixes(lookup, lacnt, preflen);
	/*sm_io_fprintf(smioout, "%d: %d unique prefixes\n", lookup->idx, num);*/

	def = ecalloc(num + 1, sizeof(def[0]));
	for (i = j = 0; i < num; i++) {
		size_t n;
		size_t plen;
		char *p;

		def[i].sm_magic = SM_CONF_DEF_MAGIC;

		plen = conf_prefix_len(lookup[j].kw + preflen);
		for (n = 0; j + n < lacnt
			     && conf_prefix_cmp(lookup[j].kw + preflen,
						lookup[j + n].kw + preflen,
						NULL) == 0;
		     n++);

		def[i].scd_name = p = emalloc(plen + 1);
		memcpy(p, lookup[j].kw + preflen, plen);
		p[plen] = 0;
		if (!(n == 1 && lookup[j].kw[preflen + plen] == 0)) {
			def[i].scd_type = sm_conf_type_section;
			def[i].scd_flags = SM_CONF_FLAG_ALLOW_ANY;
			def[i].scd_contents = create_def(base,
							 lookup + j, n,
							 preflen + plen + 1);
		} else {
			def[i].scd_type = sm_conf_type_string;
			def[i].scd_offset = (char*)&lookup[j].val
				            - (char*)base;
		}

		j += n;
	}
	def[i].scd_magic = SM_CONF_DEF_MAGIC;
	return def;
}

/* Read and parse configuration from file `conf_name', using definition DEF
   and data pointer DATA */
static void
read_conf(sm_conf_definition_T *def, void *data)
{
	sm_conf_T *smc;
	int err;

	smc = sm_conf_new(conf_name);
	if (smc == NULL) {
		sm_io_fprintf(smioerr, "%s: sm_conf_new=NULL, errno=%d\n",
	 		      conf_name, errno);
		exit((errno == ENOENT || errno == EPERM) ?
		     EX_USAGE : EX_SOFTWARE);
	}
	err = sm_conf_read_FILE(smc, conf_name, NULL);
	if (err != 0) {
		sm_prt_conferr(conf_name, smc, err, smioerr);
		exit(EX_USAGE);
	}

	err = sm_conf_scan(smc, def, SM_CONF_FLAG_ALLOW_ANY, data);
	if (err != 0) {
		sm_prt_conferr(conf_name, smc, err, smioerr);
		exit(EX_SOFTWARE);
	}
	sm_conf_destroy(smc);
}

/* Output flags */
#define OF_INDEX       0x01   /* Print indexes */
#define OF_NAME        0x02   /* Print names */
#define OF_EMPTY_SKIP  0x04   /* Skip empty values */
#define OF_EMPTY_ERROR 0x08   /* Exit with code 1 if any empty values found */
#define OF_QUOTE       0x10   /* Quote values */

int
main(int argc, char **argv)
{
	int c, i;
	int code = 0;
	sm_conf_definition_T *def;
	struct conf_lookup *lookup;
	int flags = 0;
	int sep = '\n';

	progname = argv[0];
	while ((c = getopt(argc, argv, "hf:einS:sq")) != EOF) {
		switch (c) {
		case 'h':
			usage();
			break;

		case 'f':
			conf_name = optarg;
			break;

		case 'e':
			flags |= OF_EMPTY_ERROR;
			break;

		case 'i':
			flags |= OF_INDEX;
			break;

		case 'n':
			flags |= OF_NAME;
			break;

		case 'S':
			sep = optarg[0];
			break;

		case 's':
			flags |= OF_EMPTY_SKIP;
			break;

		case 'q':
			flags |= OF_QUOTE;
			break;

		default:
			usage_error("unknown option: -%c\n", optopt);
		}
	}

	argc -= optind;
	argv += optind;

	if (argc == 0)
		usage_error("not enough arguments\n");

	/* Initialize the lookup array */
	lookup = ecalloc(argc, sizeof(lookup[0]));
	for (i = 0; i < argc; i++) {
		lookup[i].idx = i;
		lookup[i].kw = argv[i];
	}

	/* Sort by kw */
	qsort(lookup, argc, sizeof(lookup[0]), conf_lookup_kw_cmp);

	/* Create configuration definition and parse the config file. */
	def = create_def(lookup, lookup, argc, 0);
	read_conf(def, lookup);

	/* Restore original ordering of the entries. */
	qsort(lookup, argc, sizeof(lookup[0]), conf_lookup_idx_cmp);

	/* Output results. */
	for (i = 0; i < argc; i++) {
		if (!lookup[i].val) {
			if (flags & OF_EMPTY_ERROR)
				code = 1;
			if (flags & OF_EMPTY_SKIP)
				continue;
		}
		if (flags & OF_INDEX)
			sm_io_fprintf(smioout, "%d ", i);
		if (flags & OF_NAME)
			sm_io_fprintf(smioout, "%s ", lookup[i].kw);
		if (flags & OF_QUOTE)
			sm_io_fprintf(smioout, "\"%s\"", lookup[i].val ? lookup[i].val : "");
		else
			sm_io_fprintf(smioout, "%s", lookup[i].val ? lookup[i].val : "");
		sm_io_putc(smioout, sep);
	}
	/* Make sure output ends with a newline. */
	if (i && sep != '\n')
		sm_io_putc(smioout, '\n');
	sm_io_flush(smioout);
	exit(code);
}
