/*
 * Copyright (c) 2011 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-async-0.c,v 1.12 2014/01/30 16:02:28 ca Exp $")
#include "sm/common.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/heap.h"
#include "sm/test.h"
#include "sm/maps.h"
#include "sm/map.h"
#include "sm/mapclasses.h"
#include "sm/hostname.h"
#include "sm/bhtable.h"
#include "sm/io.h"
#include "sm/smardef.h"
#include "smar.h"
#include "reverse.h"
#include "dnsbl.h"
#include "t-init.h"
#include "log.h"

/* MUST be local only! required for test prg: it expects the debug output */
#ifdef SMAR_DEBUG
#undef SMAR_DEBUG
#endif
#define SMAR_DEBUG 1
#if SMAR_DEBUG
# undef SMAR_DEBFP
# undef SMAR_LEV_DPRINTF
# include "sm/io.h"
uint smar_debug;
# define SMAR_DEBFP	smioerr
# define SMAR_LEV_DPRINTF(lev, x)	do {	\
			if ((lev) < smar_debug)	{		\
				sm_io_fprintf x;		\
				sm_io_flush(SMAR_DEBFP);	\
			}					\
		} while (0)
#else
# define SMAR_LEV_DPRINTF(lev, x)
#endif

#if SMAR_USE_DNS && SMAR_ASYNC
/*
 *  Test program for (a subset of) access checks.
 *  This program is directly linked against the smar functions that
 *  perform the tests, it does not communicate with an smar daemon.
 *  This is done to allow for direct testing of one component instead
 *  of the significantly more complicated part of communicating with
 *  an smar daemon (which is done by other test programs).
 */

/* SM_AGAIN + 1 */
#define SM_WAITING	7

/*
 *  need another chk function that uses client ip,
 *  but after dnsbl; maybe cltname lookups:
 *  require more DNS lookup, good for testing several callbacks
 *  and their interaction.
 */

/* make this generic for all checks? */
#define SMARA_CFL_NONE	0x00
#define SMARA_CFL_A4	0x01
#define SMARA_CFL_GOT	0x02
#define SMARA_CFL_W4	0x04
#define SMARA_CFL_END	0x08

#define SMARA_CFL_DONE		0x80 /* done overall; hack, should be elsewhere */

static int Verbose = 0;
int Loglevel = 20;
static int Reqs = 0;
static pthread_mutex_t Mutex;
static char *Conf;

char *NS = "127.0.0.1";

#define SMARA_LFL_RVRS_T	0x00100000

typedef sm_ret_T (smar_chk_F)(smar_ctx_P, smar_addr_P, sm_ret_T *);
struct smar_chk_ctx_S
{
	const char  *arcc_name;
	smar_chk_F  *arcc_fct;
	/* anything else? preconditions? flags? */
};
typedef struct smar_chk_ctx_S smar_chk_ctx_T, *smar_chk_ctx_P;
static void prt_fct(smar_addr_P smar_addr, const char *name);

static sm_ret_T
t_smar_acc_chk(smar_ctx_P smar_ctx, smar_addr_P smar_addr, sm_ret_T ret, sm_ret_T res);

static sm_ret_T
t_smar_res(smar_ctx_P smar_ctx, smar_addr_P smar_addr)
{
	int r, reqs;
	bool stop;

	r = pthread_mutex_lock(&Mutex);
	SM_TEST(0 == r);
	reqs = Reqs;
	stop = 0 == Reqs;
	r = pthread_mutex_unlock(&Mutex);
	SM_TEST(0 == r);

	r = pthread_mutex_lock(&smar_addr->ara_mutex);
	SM_LOCK_OK(r);

	/* HACK!!! use flag in common field!!!! */
	smar_addr->ara_fl_blcb |= SMARA_CFL_DONE;
	smar_addr->ara_fl_rvcb |= SMARA_CFL_DONE;
	r = pthread_mutex_unlock(&smar_addr->ara_mutex);
	SM_UNLOCK_OK(r);

	sm_io_fprintf(smioout, "func=t_smar_res, status=%d, stop=%d\n", smar_addr->ara_status, stop);

	/* HACK */
	if (stop) {
		sm_io_flush(smioout);
		exit(0);
	}
	sm_io_fprintf(smioout, "func=t_smar_res, stop=%d, reqs=%d, cfl=%#x\n", stop, reqs, smar_addr->ara_fl_blcb); sm_io_flush(smioout);
	return SM_SUCCESS;
}

/*
**  T_ASYNC_0_INIT -- Initialize smar_ctx
**
**	Parameters:
**		smar_ctx -- SMAR context
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T smar_dns_start(smar_ctx_P smar_ctx);

static sm_ret_T
t_async_0_init(smar_ctx_P smar_ctx)
{
	sm_ret_T ret;

	ret = smar_init0(smar_ctx);
	SM_TEST_E(ret == SM_SUCCESS);

	/* initialize event threads system */
	smar_ctx->smar_max_thrds_s = 2;
	smar_ctx->smar_max_thrds_h = smar_ctx->smar_max_thrds_s * 2;
	if (smar_ctx->smar_max_thrds_s > smar_ctx->smar_max_thrds_h)
		smar_ctx->smar_max_thrds_h = UINT_MAX;
	ret = evthr_init(&smar_ctx->smar_ev_ctx, 1, smar_ctx->smar_max_thrds_s, 2);
	SM_TEST_E(ret == SM_SUCCESS);
	ret = evthr_set_max_h(smar_ctx->smar_ev_ctx, smar_ctx->smar_max_thrds_h);
	SM_TEST_E(ret == SM_SUCCESS);

	smar_ctx->smar_dns_ntsks = 1;
	smar_ctx->smar_cnf.smar_cnf_dns_flags = 0;
	ret = sm_inet_a2ipv4(NS, NULL, &smar_ctx->smar_nameserveripv4s[0]);
	SM_TEST_E(ret == SM_SUCCESS);

	smar_ctx->smar_cnf.smar_cnf_conffile = Conf;
	ret = smar_read_cnf(smar_ctx, Conf, &smar_ctx->smar_cnf.smar_cnf_smc);
	SM_TEST_E(sm_is_ok(ret));

	ret = smar_init1(smar_ctx);
	SM_TEST_E(ret == SM_SUCCESS);

	ret = pthread_mutex_init(&Mutex, NULL);
	SM_TEST(0 == ret);

	ret = smar_dns_start(smar_ctx);
	SM_TEST_E(ret == SM_SUCCESS);

  error:
	return ret;
}

/*
**  T_ASYNC_0_FREE -- Free data in smar_ctx
**
**	Parameters:
**		smar_ctx -- SMAR context
**
**	Returns:
**		usual sm_error code
*/

static void
t_async_0_free(smar_ctx_P smar_ctx)
{
	if (smar_ctx->smar_maps != NULL) {
		(void) sm_maps_term(smar_ctx->smar_maps);
		smar_ctx->smar_maps = NULL;
	}
	pthread_mutex_destroy(&Mutex);
}

/*
**  T_SMAR_CHK_DNSBL -- check DNSBL results (look up in access map)
**
**	Parameters:
**		smar_ctx -- SMAR context
**		smar_addr -- address context
**		perr -- (pointer to) error code (output)
**
**	Returns:
**		usual sm_error code
**
**	Locking: smar_addr must be under control of the caller.
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
t_smar_chk_dnsbl(smar_ctx_P smar_ctx, smar_addr_P smar_addr, sm_ret_T *perr)
{
	sm_ret_T ret;
	int r;
	char *dnsbltag;

	SM_REQUIRE(perr != NULL);
	*perr = SM_SUCCESS;

	if (!SMARA_IS_LFLAG(smar_addr, SMARA_LFL_DNSBL)) {
		SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_chk_dnsbl, status=nothing to do\n"));
		return t_smar_acc_chk(smar_ctx, smar_addr, SM_NOTDONE, SM_SUCCESS);
	}

	r = pthread_mutex_lock(&smar_addr->ara_mutex);
	SM_LOCK_OK(r);
	SM_REQUIRE((smar_addr->ara_fl_blcb & SMARA_CFL_A4) != 0);

	if ((smar_addr->ara_fl_blcb & SMARA_CFL_GOT) == 0) {
		SMAR_LEV_DPRINTF(5, (SMAR_DEBFP, "func=t_smar_chk_dnsbl, status=GOT not set\n"));
		smar_addr->ara_fl_blcb |= SMARA_CFL_W4;
		r = pthread_mutex_unlock(&smar_addr->ara_mutex);
		SM_UNLOCK_OK(r);
		return SM_WAITING;
	}

	if (smar_addr->ara_c_dnsbl >= SM_DNSBL_MAX
	    || smar_ctx->smar_cnf.smar_cnf_dnsbl[smar_addr->ara_c_dnsbl].scdb_name
	       == NULL)
	{
		SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_chk_dnsbl, flags=%#x, lflags=%#x, cdnsbl=%d, status=notdone\n", smar_addr->ara_flags, smar_addr->ara_lflags, smar_addr->ara_c_dnsbl));

		smar_addr->ara_fl_blcb |= SMARA_CFL_END;
		r = pthread_mutex_unlock(&smar_addr->ara_mutex);
		SM_UNLOCK_OK(r);
		return t_smar_acc_chk(smar_ctx, smar_addr, SM_NOTDONE, *perr);
	}

	ret = SM_SUCCESS;

	sm_str_clr(smar_addr->ara_str);
	dnsbltag = smar_ctx->smar_cnf.smar_cnf_dnsbl[smar_addr->ara_c_dnsbl].scdb_tag;
	if (NULL == dnsbltag)
		dnsbltag = "dnsbl";
	if (smar_addr->ara_dnsblres[smar_addr->ara_c_dnsbl].sbdr_res == SM_DNSBL_RES_OK
	    && sm_strprintf(smar_addr->ara_str, "%s:%A", dnsbltag,
	         smar_addr->ara_dnsblres[smar_addr->ara_c_dnsbl].sdbr_ipv4) > 0)
	{
		/* should this use _lookup_ip? */
		sm_str_clr(smar_addr->ara_rhs);
		ret = sm_map_lookup(smar_ctx->smar_access,
			SMMAPF_FL_LWR_KEY, smar_addr->ara_str, smar_addr->ara_rhs);
		SMARA_SET_FLAG(smar_addr, SMARA_FL_DIDLOOKUP);
		*perr = SM_AGAIN;
		SMAR_LEV_DPRINTF(5, (SMAR_DEBFP, "func=t_smar_chk_dnsbl, dnsbl=%S, ret=%r, rhs=%#S\n", smar_addr->ara_str, ret, smar_addr->ara_rhs));

		/*
		 *  check the rest... but how to get the result back?
		 *  add a counter:
		 *  smar_addr->ara_dnsblres[smar_addr->ara_c_dnsbl].sbdr_i
		 *  and check it against the limit in the statement above similar to
		 *  smar_addr->ara_c_dnsbl >= SM_DNSBL_MAX
		 *
		 *  however, then we still have the problem of selecting the "right"
		 *  result, e.g., perm over temp error.
		 *
		 *  call something like "analyze" to extract result?
		 */

		/* hack for now: try to find the first match, this is random! */
		if (smar_addr->ara_dnsblres[smar_addr->ara_c_dnsbl].sdbr_n > 0 &&
			smar_addr->ara_dnsblres[smar_addr->ara_c_dnsbl].sdbr_ipv4s != NULL
			&& ret != SM_SUCCESS
		   )
		{
			uint ui;
			ipv4_T *ipv4s;

			for (ui = 0,
				 ipv4s = smar_addr->ara_dnsblres[smar_addr->ara_c_dnsbl].sdbr_ipv4s;
				 ui < smar_addr->ara_dnsblres[smar_addr->ara_c_dnsbl].sdbr_n &&
				 ret != SM_SUCCESS;
				 ++ui)
			{
				sm_str_clr(smar_addr->ara_str);
				if (sm_strprintf(smar_addr->ara_str, "%s:%A", dnsbltag,
						ipv4s[ui]) > 0)
				{
					sm_str_clr(smar_addr->ara_rhs);
					ret = sm_map_lookup(smar_ctx->smar_access,
						SMMAPF_FL_LWR_KEY, smar_addr->ara_str,
						smar_addr->ara_rhs);
					SMAR_LEV_DPRINTF(5, (SMAR_DEBFP, "func=t_smar_chk_dnsbl, dnsbl=%S, ret=%r, rhs=%#S\n", smar_addr->ara_str, ret, smar_addr->ara_rhs));
				}
			}
		}

	}
	else if (smar_addr->ara_dnsblres[smar_addr->ara_c_dnsbl].sbdr_res == DNSR_TEMP
	    && sm_str_scatv(smar_addr->ara_str, 2, dnsbltag, ":temp") == SM_SUCCESS)
	{
		sm_str_clr(smar_addr->ara_rhs);
		ret = sm_map_lookup(smar_ctx->smar_access,
			SMMAPF_FL_LWR_KEY, smar_addr->ara_str, smar_addr->ara_rhs);
		SMARA_SET_FLAG(smar_addr, SMARA_FL_DIDLOOKUP);
		*perr = SM_AGAIN;
		SMAR_LEV_DPRINTF(5, (SMAR_DEBFP, "func=t_smar_chk_dnsbl, dnsbl=%S, ret=%r, rhs=%#S\n", smar_addr->ara_str, ret, smar_addr->ara_rhs));
	}
	else if (smar_addr->ara_dnsblres[smar_addr->ara_c_dnsbl].sbdr_res == DNSR_PERM
	    && sm_str_scatv(smar_addr->ara_str, 2, dnsbltag, ":perm") == SM_SUCCESS)
	{
		sm_str_clr(smar_addr->ara_rhs);
		ret = sm_map_lookup(smar_ctx->smar_access,
			SMMAPF_FL_LWR_KEY, smar_addr->ara_str, smar_addr->ara_rhs);
		SMARA_SET_FLAG(smar_addr, SMARA_FL_DIDLOOKUP);
		*perr = SM_AGAIN;
		SMAR_LEV_DPRINTF(5, (SMAR_DEBFP, "func=t_smar_chk_dnsbl, dnsbl=%S, ret=%r, rhs=%#S\n", smar_addr->ara_str, ret, smar_addr->ara_rhs));
	}
	++smar_addr->ara_c_dnsbl;
	r = pthread_mutex_unlock(&smar_addr->ara_mutex);
	SM_UNLOCK_OK(r);
	return t_smar_acc_chk(smar_ctx, smar_addr, ret, *perr);
}

/*
**  T_SMAR_DNSBL_NOTIFY -- callback for a DNS BL query
**
**	Parameters:
**		smar_dnsbl -- DNS BL context
**		ctx -- smar_addr
**
**	Returns:
**		usual sm_error code
**
**	Locking: locks smar_addr
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
t_smar_dnsbl_notify(smar_dnsbl_P smar_dnsbl, void *ctx)
{
	sm_ret_T ret;
	int r;
	bool stop, analyze;
	smar_addr_P smar_addr;

	SM_REQUIRE(smar_dnsbl != NULL);
	SM_REQUIRE(ctx != NULL);
	smar_addr = ctx;
	SM_IS_SMAR_ADDR(smar_addr);
	ret = SM_SUCCESS;
	stop = analyze = false;

	r = pthread_mutex_lock(&Mutex);
	SM_TEST(0 == r);
	SM_TEST(Reqs > 0);
	--Reqs;
	r = pthread_mutex_unlock(&Mutex);
	SM_TEST(0 == r);

	SMAR_LEV_DPRINTF(9, (SMAR_DEBFP, "func=t_smar_dnsbl_notify, smar_addr=%p\n", smar_addr));
	r = pthread_mutex_lock(&smar_addr->ara_mutex);
	SM_LOCK_OK(r);
	SM_REQUIRE((smar_addr->ara_fl_blcb & SMARA_CFL_A4) != 0);

	SM_ASSERT(smar_addr->ara_res_cnt > 0);
	--smar_addr->ara_res_cnt;
	if (0 == smar_addr->ara_res_cnt) {
		smar_addr->ara_fl_blcb |= SMARA_CFL_GOT;

		/* only invoke dnsbl if we aren't done yet */
		if (0 == (smar_addr->ara_fl_blcb & SMARA_CFL_DONE))
			analyze = (smar_addr->ara_fl_blcb & SMARA_CFL_W4) != 0;
		else
			stop = true;
	}

	SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_dnsbl_notify, res_cnt=%d, res=%r, ipv4=%A, analyze=%d, cfl=%#x\n", smar_addr->ara_res_cnt, smar_dnsbl->adbl_res->sbdr_res, smar_dnsbl->adbl_res->sdbr_ipv4, analyze, smar_addr->ara_fl_blcb));

	r = pthread_mutex_unlock(&smar_addr->ara_mutex);
	SM_UNLOCK_OK(r);

	if (analyze) {
		/*
		 *  how to continue properly?
		 *  should t_smar_acc_chk() check what to do?
		 *  however, t_smar_chk_dnsbl() was already called and the
		 *  fct_idx is "beyond" it, so this seems to be the right call.
		 */
		prt_fct(smar_addr, "t_smar_dnsbl_notify");
/*
		return t_smar_chk_dnsbl(smar_addr->ara_smar_clt_ctx->smac_ar_ctx, smar_addr, &ret);
*/
		return t_smar_acc_chk(smar_addr->ara_smar_clt_ctx->smac_ar_ctx, smar_addr, ret, SM_AGAIN);
	}
	if (stop) {
		/* HACK */
		SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_dnsbl_notify, stop=%d\n", stop));
		exit(0);
	}
	return SM_SUCCESS;
}

/*
**  T_SMAR_START_RV -- start reverse lookup
**
**	Parameters:
**		smar_ctx -- SMAR context
**		smar_addr -- address context
**		perr -- (pointer to) error code (output)
**
**	Returns:
**		usual sm_error code
**
**	Last code review:
**	Last code change:
*/

extern sm_ret_T smar_rvrs_cb(dns_res_P dns_res, void *ctx);

static sm_ret_T
t_smar_start_rv(smar_ctx_P smar_ctx, smar_addr_P smar_addr, sm_ret_T *perr)
{
	sm_ret_T ret;
	int r;
	dns_mgr_ctx_P dns_mgr_ctx;
	sm_cstr_P q;
	smar_rvrs_P smar_rvrs;

	SM_REQUIRE(perr != NULL);
	*perr = SM_SUCCESS;

	if (!SMARA_IS_LFLAG(smar_addr, SMARA_LFL_RVRS_T)) {
		SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_start_rv, status=nothing to do\n"));
		return t_smar_acc_chk(smar_ctx, smar_addr, SM_NOTDONE, SM_SUCCESS);
	}
	SM_IS_SMAR_ADDR(smar_addr);
	SM_IS_SMAR_CTX(smar_ctx);

	r = pthread_mutex_lock(&smar_addr->ara_mutex);
	SM_LOCK_OK(r);

	SM_ASSERT(0 == smar_addr->ara_res_cnt);
	q = NULL;
	dns_mgr_ctx = smar_ctx->smar_dns_mgr_ctx;
	smar_rvrs = smar_addr->ara_rvrs;
	smar_rvrs->arv_pa = sm_str_new(NULL, IP_ARPASTRLEN, IP_ARPASTRMAX);
	ret = sm_inet_ip2arpastr(&smar_rvrs->arv_ip, smar_rvrs->arv_pa);
	if (sm_is_err(ret))
		goto error;
	q = sm_cstr_scpyn0((const uchar *) sm_str_getdata(smar_rvrs->arv_pa),
			sm_str_getlen(smar_rvrs->arv_pa));
	if (NULL == q) {
		ret = sm_error_perm(SM_EM_AR, ENOMEM);
		goto error;
	}

	/*
	**  Set flag before passing smar_rvrs to dns_req_add()
	**  because smar_rvrs_cb() will free() smar_rvrs.
	*/

	SMARV_SET_FLAG(smar_rvrs, SMARV_FL_A4PTR);
	smar_rvrs->arv_PTR_qsent = 1;
	ret = dns_req_add(dns_mgr_ctx, q, T_PTR, 0, smar_rvrs_cb, smar_rvrs);
	/* XXX check return? */
	SM_TEST_E(!sm_is_err(ret));

	smar_addr->ara_fl_rvcb |= SMARA_CFL_A4;
	++smar_addr->ara_res_cnt;
	SMARA_SET_FLAG(smar_addr, SMARA_FL_A4RVRS);
	r = pthread_mutex_unlock(&smar_addr->ara_mutex);
	SM_UNLOCK_OK(r);

	SM_CSTR_FREE(q);
	SMAR_LEV_DPRINTF(6, (SMAR_DEBFP, "func=t_smar_start_rv, addr=%#S\n", smar_rvrs->arv_pa));
	r = pthread_mutex_lock(&Mutex);
	SM_TEST(0 == r);
	++Reqs;
	r = pthread_mutex_unlock(&Mutex);
	SM_TEST(0 == r);
	return t_smar_acc_chk(smar_ctx, smar_addr, ret, SM_SUCCESS);

  error:
	r = pthread_mutex_unlock(&smar_addr->ara_mutex);
	SM_UNLOCK_OK(r);
	sm_log_write(smar_ctx->smar_lctx,
		AR_LCAT_RESOLVER, AR_LMOD_RESOLVER, SM_LOG_ERROR, 0,
		"sev=ERROR, func=smar_rvrs_rslv, ret=%x", ret);
	return t_smar_acc_chk(smar_ctx, smar_addr, ret, SM_SUCCESS);
}

/*
**  T_SMAR_START_DNSBL -- check DNSBL results (look up in access map)
**
**	Parameters:
**		smar_ctx -- SMAR context
**		smar_addr -- address context
**		perr -- (pointer to) error code (output)
**
**	Returns:
**		usual sm_error code
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
t_smar_start_dnsbl(smar_ctx_P smar_ctx, smar_addr_P smar_addr, sm_ret_T *perr)
{
	sm_ret_T ret;
	int r;
	ipv4_T ipv4;
	uint u;
	smar_dnsbl_P smar_dnsbl;

	SM_REQUIRE(perr != NULL);
	*perr = SM_SUCCESS;

	if (!SMARA_IS_LFLAG(smar_addr, SMARA_LFL_DNSBL)) {
		SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_start_dnsbl, status=nothing to do\n"));
		return t_smar_acc_chk(smar_ctx, smar_addr, SM_NOTDONE, SM_SUCCESS);
	}
	SM_IS_SMAR_ADDR(smar_addr);
	SM_IS_SMAR_CTX(smar_ctx);

	r = pthread_mutex_lock(&smar_addr->ara_mutex);
	SM_LOCK_OK(r);

	/* all flags for this op must be cleared */
	SM_REQUIRE(smar_addr->ara_fl_blcb == 0);
	smar_addr->ara_fl_blcb |= SMARA_CFL_A4;
	/*SM_ASSERT(0 == smar_addr->ara_res_cnt);*/

	ret = SM_SUCCESS;
	for (u = 0;
	     u < SM_DNSBL_MAX &&
	     smar_ctx->smar_cnf.smar_cnf_dnsbl[u].scdb_name != NULL;
	     u++)
	{
		SMAR_LEV_DPRINTF(11, (SMAR_DEBFP, "func=t_smar_start_dnsbl, u=%u\n", u));
		ret = smar_dnsbl_new(smar_addr, &(smar_addr->ara_dnsblres[u]),
				&(smar_ctx->smar_cnf.smar_cnf_dnsbl[u]),
				t_smar_dnsbl_notify, smar_addr, &smar_dnsbl);
		SMAR_LEV_DPRINTF(9, (SMAR_DEBFP, "func=t_smar_start_dnsbl, u=%u, smar_dnsbl_new=%r\n", u, ret));
		if (sm_is_err(ret))
			goto unlock;	/* XXX set some error code, log? */

		r = pthread_mutex_lock(&Mutex);
		SM_TEST(0 == r);
		++Reqs;
		r = pthread_mutex_unlock(&Mutex);
		SM_TEST(0 == r);

		++smar_addr->ara_res_cnt;
		ret = sm_inet_a2ipv4((const char *)sm_str_getdata(smar_addr->ara_pa),
			NULL, &ipv4);

		/*
		**  This is an asynchronous function!
		**  Note: smar_dnsbl is now under control of smar_dnsbl_rslv().
		*/

		ret = smar_dnsbl_rslv(smar_ctx, smar_dnsbl, ipv4, 0u);
		if (sm_is_success(ret))
			SMARA_SET_FLAG(smar_addr, SMARA_FL_A4DNSBL);
		else {
			--smar_addr->ara_res_cnt;
			ret = SM_SUCCESS; /* don't return error to caller */
		}
		SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "sev=DBG, func=smar_startdnsbl, smar_dnsbl_rslv=%r, ipv4=%A, scdb_name=%s\n", ret, ipv4, smar_ctx->smar_cnf.smar_cnf_dnsbl[u].scdb_name));
	}

	/* ignore error, ara_status is initialized */

  unlock:
	r = pthread_mutex_unlock(&smar_addr->ara_mutex);
	SM_UNLOCK_OK(r);
	SMAR_LEV_DPRINTF(19, (SMAR_DEBFP, "func=t_smar_start_dnsbl, endu=%u\n", u));
	return t_smar_acc_chk(smar_ctx, smar_addr, ret, SM_SUCCESS);
}

/*
**  T_SMAR_CHK_IPV4 -- perform IPv4 access map lookup.
**
**	Parameters:
**		smar_ctx -- SMAR context
**		smar_addr -- address context
**		perr -- (pointer to) error code (output)
**
**	Returns:
**		usual sm_error code
**
**	Locking: smar_addr must be under control of the caller.
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
t_smar_chk_ipv4(smar_ctx_P smar_ctx, smar_addr_P smar_addr, sm_ret_T *perr)
{
	sm_ret_T ret;

	SM_IS_SMAR_ADDR(smar_addr);
	SM_IS_SMAR_CTX(smar_ctx);
	SM_REQUIRE(perr != NULL);
	*perr = SM_SUCCESS;

	if (!SMARA_IS_LFLAG(smar_addr, SMARA_LFL_IPV4)
	    || !SMARA_IS_LFLAG(smar_addr, SMARA_LFL_SUB)
	    || SMARA_IS_FLAG(smar_addr, SMARA_FL_GOTIPV4ACC))
	{
		SMAR_LEV_DPRINTF(3, (SMAR_DEBFP, "func=t_smar_chk_ipv4, lfl=%#x, status=nothing to do\n", smar_addr->ara_lflags));
		return t_smar_acc_chk(smar_ctx, smar_addr, SM_NOTDONE, *perr);
	}

	sm_str_clr(smar_addr->ara_tag);
	(void) sm_str_scat(smar_addr->ara_tag, CLT_A_TAG);
	ret = sm_map_lookup_ip(smar_ctx->smar_access,
			smar_addr->ara_str, smar_addr->ara_tag,
			SMMAP_LFL_SUBNETS|SMMAP_LFL_TAG, smar_addr->ara_rhs);
	SMARA_SET_FLAG(smar_addr, SMARA_FL_DIDLOOKUP|SMARA_FL_GOTIPV4ACC);
	SMAR_LEV_DPRINTF(3, (SMAR_DEBFP, "sev=DBG, func=t_smar_chk_ipv4, ip=%S, tag=%S, ret=%r\n", smar_addr->ara_str, smar_addr->ara_tag, ret));
	return t_smar_acc_chk(smar_ctx, smar_addr, ret, *perr);
}

/*
**  T_SMAR_CLTNAME -- Perform clientname check
**
**	Parameters:
**		smar_ctx -- SMAR context
**		smar_addr -- address context
**		perr -- (pointer to) error code (output)
**
**	Returns:
**		usual sm_error code
**
**	Locking: smar_addr must be under control of the caller.
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
t_smar_cltname(smar_ctx_P smar_ctx, smar_addr_P smar_addr, sm_ret_T *perr)
{
	sm_ret_T ret;
	int r;

	SM_REQUIRE(perr != NULL);
	ret = *perr = SM_SUCCESS;

	if (!SMARA_IS_LFLAG(smar_addr, SMARA_LFL_RVACC)) {
		SMAR_LEV_DPRINTF(3, (SMAR_DEBFP, "func=t_smar_cltname, status=nothing to do\n"));
		return t_smar_acc_chk(smar_ctx, smar_addr, SM_NOTDONE, SM_SUCCESS);
	}

	r = pthread_mutex_lock(&smar_addr->ara_mutex);
	SM_LOCK_OK(r);

	if ((smar_addr->ara_fl_rvcb & SMARA_CFL_GOT) == 0) {
		SMAR_LEV_DPRINTF(5, (SMAR_DEBFP, "func=t_smar_cltname, status=GOT not set\n"));
		smar_addr->ara_fl_rvcb |= SMARA_CFL_W4;
		r = pthread_mutex_unlock(&smar_addr->ara_mutex);
		SM_UNLOCK_OK(r);
		return SM_WAITING;
	}

	SMAR_LEV_DPRINTF(8, (SMAR_DEBFP, "func=t_smar_cltname, hostname=%#C\n", smar_addr->ara_rvrs_hostname));
	if (NULL != smar_addr->ara_rvrs_hostname) {
		sm_str_clr(smar_addr->ara_tag);
		ret = sm_str_scat(smar_addr->ara_tag, CLT_N_TAG);
		ret = sm_map_lookup_domain(smar_ctx->smar_access,
			(sm_rdstr_P) smar_addr->ara_rvrs_hostname,
			smar_addr->ara_tag, 0, smar_addr->ara_rhs);
		SMARA_SET_FLAG(smar_addr, SMARA_FL_DIDLOOKUP|SMARA_FL_GOTRVRSN);
	}

	r = pthread_mutex_unlock(&smar_addr->ara_mutex);
	SM_UNLOCK_OK(r);
	SMAR_LEV_DPRINTF(3, (SMAR_DEBFP, "func=t_smar_cltname, hostname=%#C, ret=%r\n", smar_addr->ara_rvrs_hostname, ret));
	return t_smar_acc_chk(smar_ctx, smar_addr, ret, *perr);
}

/*
**  T_SMAR_CLTRESOLVE -- Perform cltresolve check
**
**	Parameters:
**		smar_ctx -- SMAR context
**		smar_addr -- address context
**		perr -- (pointer to) error code (output)
**
**	Returns:
**		usual sm_error code
**
**	Locking: smar_addr must be under control of the caller.
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
t_smar_cltresolve(smar_ctx_P smar_ctx, smar_addr_P smar_addr, sm_ret_T *perr)
{
	sm_ret_T ret;
	int r;
#define CLT_RVRS_TAG	"cltresolve:"

	SM_REQUIRE(perr != NULL);
	ret = *perr = SM_SUCCESS;

	/* should this function be used at all? */
	if (!SMARA_IS_LFLAG(smar_addr, SMARA_LFL_RVACC)) {
		SMAR_LEV_DPRINTF(3, (SMAR_DEBFP, "func=t_smar_cltresolve, status=nothing to do\n"));
		return t_smar_acc_chk(smar_ctx, smar_addr, SM_NOTDONE, SM_SUCCESS);
	}

	r = pthread_mutex_lock(&smar_addr->ara_mutex);
	SM_LOCK_OK(r);

	/* do we have the data that we need? */
	if ((smar_addr->ara_fl_rvcb & SMARA_CFL_GOT) == 0) {
		SMAR_LEV_DPRINTF(5, (SMAR_DEBFP, "func=t_smar_cltresolve, status=GOT not set\n"));
		smar_addr->ara_fl_rvcb |= SMARA_CFL_W4;
		r = pthread_mutex_unlock(&smar_addr->ara_mutex);
		SM_UNLOCK_OK(r);
		return SM_WAITING;
	}

	/* perform the operation */
	sm_str_clr(smar_addr->ara_str);
	if (sm_str_scat(smar_addr->ara_str, CLT_RVRS_TAG) == SM_SUCCESS
	    && sm_str_scat(smar_addr->ara_str,
		sm_rslv2txt(smar_addr->ara_rvrs->arv_ret)) == SM_SUCCESS)
	{
		sm_str_clr(smar_addr->ara_rhs);
		ret = sm_map_lookup(smar_ctx->smar_access, SMMAPF_FL_LWR_KEY,
				smar_addr->ara_str, smar_addr->ara_rhs);
		SMARA_SET_FLAG(smar_addr, SMARA_FL_DIDLOOKUP);
	}
	SMARA_SET_FLAG(smar_addr, SMARA_FL_GOTRVRSA);

	r = pthread_mutex_unlock(&smar_addr->ara_mutex);
	SM_UNLOCK_OK(r);
	SMAR_LEV_DPRINTF(3, (SMAR_DEBFP, "func=t_smar_cltresolve, ret=%r\n", ret));
	return t_smar_acc_chk(smar_ctx, smar_addr, ret, *perr);
}

/*
**  T_SMAR_RVRS_NOTIFY -- callback for a reverse resolution query
**
**	Parameters:
**		smar_rvrs -- reverse resolution context
**		ctx -- smar_addr
**
**	Returns:
**		usual sm_error code
**
**	Locking: locks smar_addr
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
t_smar_rvrs_notify(smar_rvrs_P smar_rvrs, void *ctx)
{
	sm_ret_T ret;
	int r, rescnt;
	bool stop, analyze;
	smar_addr_P smar_addr;

/* See t_smar_dnsbl_notify(); turn it into a generic function? */

	SM_REQUIRE(smar_rvrs != NULL);
	SM_REQUIRE(ctx != NULL);
	smar_addr = ctx;
	ret = SM_SUCCESS;
	stop = analyze = false;

	SMAR_LEV_DPRINTF(8, (SMAR_DEBFP, "func=t_smar_rvrs_notify\n"));

	r = pthread_mutex_lock(&Mutex);
	SM_TEST(0 == r);
	SM_TEST(Reqs > 0);
	--Reqs;
	r = pthread_mutex_unlock(&Mutex);
	SM_TEST(0 == r);

	r = pthread_mutex_lock(&smar_addr->ara_mutex);
	SM_LOCK_OK(r);
	if (smar_rvrs->arv_hostname != NULL)
		smar_addr->ara_rvrs_hostname = SM_CSTR_DUP(smar_rvrs->arv_hostname);
	SM_REQUIRE(smar_rvrs->arv_ret != 0);

	SM_REQUIRE((smar_addr->ara_fl_rvcb & SMARA_CFL_A4) != 0);

	smar_addr->ara_dns_ret = smar_rvrs->arv_ret;
	SM_ASSERT(smar_addr->ara_res_cnt > 0);
	rescnt = --smar_addr->ara_res_cnt;

	if (0 == smar_addr->ara_res_cnt) {
		smar_addr->ara_fl_rvcb |= SMARA_CFL_GOT;

		/* only invoke t_smar_cltname if we aren't done yet */
		if (0 == (smar_addr->ara_fl_rvcb & SMARA_CFL_DONE))
			analyze = (smar_addr->ara_fl_rvcb & SMARA_CFL_W4) != 0;
		else
			stop = true;
	}

	r = pthread_mutex_unlock(&smar_addr->ara_mutex);
	SM_UNLOCK_OK(r);

SMAR_LEV_DPRINTF(8, (SMAR_DEBFP, "func=t_smar_rvrs_notify, rescnt=%d, ret=%m, analyze=%d\n", rescnt, smar_addr->ara_dns_ret, analyze));
	if (analyze) {
		/*
		 *  how to continue properly?
		 */
		prt_fct(smar_addr, "t_smar_rvrs_notify");
/*
		return t_smar_cltname(smar_addr->ara_smar_clt_ctx->smac_ar_ctx, smar_addr, &ret);
*/
		return t_smar_acc_chk(smar_addr->ara_smar_clt_ctx->smac_ar_ctx, smar_addr, ret, SM_AGAIN);
	}
	if (stop) {
		/* HACK */
		SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_rvrs_notify, stop=%d\n", stop));
		exit(0);
	}
	return SM_SUCCESS;
}

const smar_chk_ctx_T t_ar_chks[] = {
	{ "t_smar_start_rv",	t_smar_start_rv },
	{ "t_smar_chk_ipv4",	t_smar_chk_ipv4 },
	{ "t_smar_cltresolve",  t_smar_cltresolve },
	{ "t_smar_cltname",		t_smar_cltname },
	{ "t_smar_start_dnsbl",	t_smar_start_dnsbl },
	{ "t_smar_chk_dnsbl",	t_smar_chk_dnsbl  },
/*	{ "t_smar_mxbadip",		t_smar_mxbadip },	*/
/*	{ "t_smar_res",			t_smar_res }, */
	{ NULL,					(smar_chk_F*)NULL},
};

static void
prt_fct(smar_addr_P smar_addr, const char *name)
{
	if (smar_addr->ara_fct_idx < SM_ARRAY_SIZE(t_ar_chks)) {
SMAR_LEV_DPRINTF(8, (SMAR_DEBFP, "func=%s, fct_idx=%d, fct=%s\n", name, smar_addr->ara_fct_idx, t_ar_chks[smar_addr->ara_fct_idx].arcc_name));
	}
}

/*
**  T_SMAR_ACC_CHK -- Check address against access map and perform other
**		checks as requested (e.g., DNS)
**
**	Parameters:
**		smar_ctx -- SMAR context
**		smar_addr -- address context
**
**	Returns:
**		usual sm_error code
**		if ok: smar_addr->ara_status should have result of lookup(s)
**
**	Locking: smar_addr must be under control of the caller.
**
**	Last code review:
**	Last code change:
**
**	split this into smaller functions that can be called individually?
*/

static sm_ret_T
t_smar_acc_chk(smar_ctx_P smar_ctx, smar_addr_P smar_addr, sm_ret_T ret, sm_ret_T res)
{
	smar_chk_F *smar_cc_fct;

	SM_IS_SMAR_ADDR(smar_addr);
	SM_IS_SMAR_CTX(smar_ctx);

	/* ret = (*smar_cc_fct)(smar_ctx, smar_addr, &res); */

	SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_acc_chk, ret=%r, res=%r, flags=%#x\n", ret, res, smar_addr->ara_flags));
	if (sm_is_err(res)) return ret; /* res? */
	/* if (SM_BREAK == res) break; ??? only one fct uses it */
	if (SM_DONE == res) return SM_SUCCESS;
	if (res != SM_AGAIN) ++smar_addr->ara_fct_idx;

	if (SMARA_IS_FLAG(smar_addr, SMARA_FL_DIDLOOKUP)) {
		ret = smar_acc_analyse(smar_ctx, smar_addr, ret);
		SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_acc_chk, smar_acc_analyse=%r, status=%d\n", ret, smar_addr->ara_status));
		if (SMARA_IS_FLAG(smar_addr, SMARA_FL_RP_ANALYSE))
			SMARA_CLR_FLAG(smar_addr, SMARA_FL_RP_ANALYSE);
		if (SMARA_IS_FLAG(smar_addr, SMARA_FL_RC_ANALYSE))
			SMARA_CLR_FLAG(smar_addr, SMARA_FL_RC_ANALYSE);
		if (!sm_is_err(ret))
			SMARA_CLR_FLAG(smar_addr, SMARA_FL_DIDLOOKUP);
	}

	SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_acc_chk, flags=%#x, status=%d\n", smar_addr->ara_flags, smar_addr->ara_status));
	if (smar_addr->ara_status == SMTP_R_CONT
       && !SMARA_IS_FLAG(smar_addr, SMARA_FL_STOP)
       && !SMARA_IS_FLAG(smar_addr, SMARA_FL_DIDLOOKUP)
       && smar_addr->ara_fct_idx < SM_ARRAY_SIZE(t_ar_chks)
	   && (smar_cc_fct = t_ar_chks[smar_addr->ara_fct_idx].arcc_fct) != NULL)
	{
		/*
		**  ret is used by smar_acc_analyse()
		**  res is used "locally" to make decision about the control flow,
		**  e.g., whether to "error out", break, etc.
		*/

		SMAR_LEV_DPRINTF(11, (SMAR_DEBFP, "func=t_smar_acc_chk, fct=%p\n", smar_cc_fct));
		return (*smar_cc_fct)(smar_ctx, smar_addr, &res);
	}
	else	/* make this simply the last entry in the array? */
		return t_smar_res(smar_ctx, smar_addr);
}

static sm_ret_T
t_smar_access_chk(smar_ctx_P smar_ctx, smar_addr_P smar_addr)
{
	sm_ret_T ret, res;

	SM_IS_SMAR_ADDR(smar_addr);
	SM_IS_SMAR_CTX(smar_ctx);

	/* move to smar_access_check()??? only if the cleanup code is moved too */
	ret = smar_addr_init(smar_ctx, smar_addr, &res);
	SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "smar_addr_init=%r, res=%r\n", ret, res));
	SM_TEST_E(!sm_is_err(res));
	if (sm_is_err(res))
		goto error;
	if (SM_DONE == res)
		goto done;
	SM_TEST_E(smar_addr->ara_str != NULL);

	/* copy from smar_addr_init() without smar_rvrs_rslv() */
	if (SMARA_IS_LFLAG(smar_addr, SMARA_LFL_RVRS_T)) {
		ret = smar_rvrs_new(smar_ctx, smar_addr->ara_smar_clt_ctx, &smar_addr->ara_rvrs);
		SM_TEST_E(!sm_is_err(ret));
		if (sm_is_err(ret))
			goto error;	/* XXX set some error code, log? */

		SMARV_SET_FLAG(smar_addr->ara_rvrs, SMARV_FL_NOSEND|SMARV_FL_NOFREE);
		SESSTA_COPY(smar_addr->ara_rvrs->arv_seid, smar_addr->ara_taid);
		smar_addr->ara_rvrs->arv_ltype = smar_addr->ara_ltype;
		smar_addr->ara_rvrs->arv_lflags = smar_addr->ara_lflags;
		ret = sm_inet_a2ip((const char *)sm_str_getdata(smar_addr->ara_pa),
			NULL, &smar_addr->ara_rvrs->arv_ip);
SMAR_LEV_DPRINTF(4, (SMAR_DEBFP, "sm_inet_a2ip=%r, pa=%#S\n", ret, smar_addr->ara_pa));
		smar_addr->ara_rvrs->arv_cbf = t_smar_rvrs_notify;
		smar_addr->ara_rvrs->arv_cb_ctx = smar_addr;
		/* ignore error, ara_status is initialized */
	}

	return t_smar_acc_chk(smar_ctx, smar_addr, SM_SUCCESS, SM_AGAIN);

  done:

	/* claim that everything is ok */
	ret = SM_SUCCESS;
	/* fallthrough for cleanup */

  error:
	if (smar_addr->ara_rhs != NULL && sm_str_getlen(smar_addr->ara_rhs) > 0) {
		ret = sm_str_scat(smar_addr->ara_rhs, "\r\n");
		SM_ASSERT(SM_SUCCESS == ret);
	}
	if (SMARA_IS_FLAG(smar_addr, SMARA_FL_RFC2821))
		a2821_free(&smar_addr->ara_a_rcpt);
	SM_STR_FREE(smar_addr->ara_str);
	SM_STR_FREE(smar_addr->ara_tag);
	SM_STR_FREE(smar_addr->ara_detail);
	SM_STR_FREE(smar_addr->ara_domain_pa);

	return ret;
}

static sm_ret_T
t_smar_main(sm_evthr_task_P tsk)
{
	sm_ret_T ret;
	int r, stop;
	smar_addr_P smar_addr;

	SM_IS_EVTHR_TSK(tsk);
	smar_addr = (smar_addr_P) tsk->evthr_t_actx;

	ret = t_smar_access_chk(smar_addr->ara_smar_clt_ctx->smac_ar_ctx, smar_addr);
	r = pthread_mutex_lock(&Mutex);
	SM_TEST(0 == r);
	stop = 0 == Reqs;
	r = pthread_mutex_unlock(&Mutex);
	SM_TEST(0 == r);

	SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "func=t_smar_main, ret=%r, stop=%d\n", ret, stop));

	return EVTHR_DEL|(stop ? EVTHR_TERM : 0);
}

/*
**  T_ASYNC_0 -- Perform several access check
**
**	Parameters:
**		ipaddr -- (printable) address to check.
**
**	Returns:
**		none.
*/

static void
t_async_0(const char *ipaddr, uint32_t lflags)
{
	sm_ret_T ret;
	smar_addr_P smar_addr;
	smar_ctx_T smar_ctx;
	sm_evthr_task_P task;
	ipv4_T ipv4;
	timeval_T sleept;

	ret = t_async_0_init(&smar_ctx);
	SM_TEST_E(ret == SM_SUCCESS);

	ret = smar_addr_new(smar_ctx.smar_clt_ctx[0], &smar_addr);
	SM_TEST_E(ret == SM_SUCCESS);

	ret = sm_inet_a2ipv4(ipaddr, NULL, &ipv4);
	SM_TEST_E(ret == SM_SUCCESS);

	smar_addr->ara_ltype = SMARA_LT_CLT_A_ACC;
	smar_addr->ara_lflags = lflags;
	smar_addr->ara_pa = sm_str_scpy(NULL, ipaddr, 16);
	SM_TEST_E(smar_addr->ara_pa != NULL);
	smar_addr->ara_pa2 = NULL;
	smar_addr->ara_sckaddr.sa.sa_family = AF_INET;
	smar_addr->ara_sckaddr.sin.sin_addr.s_addr = ipv4;
	SMARA_SET_FLAG(smar_addr, SMARA_FL_RCVDIP);

	ret = gettimeofday(&sleept, NULL);
	SM_TEST(0 == ret);
	sleept.tv_usec += 1;
	ret = evthr_task_new(smar_ctx.smar_ev_ctx, &task, EVTHR_EV_SL,
			-1, &sleept, t_smar_main, (void *) smar_addr);
	SM_TEST(sm_is_success(ret));
	SM_TEST(task != NULL);

	ret = evthr_loop(smar_ctx.smar_ev_ctx);
	SM_TEST_E(ret == SM_SUCCESS);

	if (ret == SM_SUCCESS) {
		sm_io_fprintf(smioout, "func=t_async_0, ara_status=%r\n", smar_addr->ara_status);
		sm_io_flush(smioout);
	}
	else if (Verbose > 0)
		SMAR_LEV_DPRINTF(7, (SMAR_DEBFP, "smar_rcpt_expand=%#X\n", ret));

	SM_TEST(ret == SM_SUCCESS);
	t_async_0_free(&smar_ctx);

  error:
	return;
}

/*
**  USAGE -- usage message
**
**	Parameters:
**		prg -- program name
**
**	Returns:
**		none
*/

static void
usage(const char *prg)
{
	sm_io_fprintf(smioerr, "usage: %s [options] addresses...\n"
		"-F n       set lflags\n"
		"-f conf    use conf\n"
#if SMAR_DEBUG
		"-d level   set debug level\n"
#endif
		"-i ns      specify nameserver\n"
		"-R         perform only reverse lookup\n"
		"-r         perform reverse lookup too\n"
		"-V         increase verbosity\n"
		"-v n       set log level\n"
		, prg
		);
	return;
}

int
main(int argc, char *argv[])
{
	int c, i;
	uint32_t lflags;

#if SMAR_DEBUG
	smar_debug = 8;
#endif
	Conf = "smar.conf";
	lflags = SMARA_LFL_IPV4|SMARA_LFL_SUB|SMARA_LFL_DNSBL;
	while ((c = getopt(argc, argv, "d:F:f:i:RrVv:")) != -1) {
		switch (c) {
		  case 'd':
#if SMAR_DEBUG
			smar_debug = (uint) strtoul(optarg, NULL, 0);
#endif
			break;
		  case 'F':
			lflags = (uint32_t) strtoul(optarg, NULL, 0);
			break;
		  case 'f':
			Conf = strdup(optarg);
			break;
		  case 'i':
			NS = strdup(optarg);
			break;
		  case 'R':
			lflags = SMARA_LFL_IPV4|SMARA_LFL_SUB|SMARA_LFL_RVACC|SMARA_LFL_RVRS_T;
			break;
		  case 'r':
			lflags |= SMARA_LFL_RVACC|SMARA_LFL_RVRS_T;
			break;
		  case 'V':
			++Verbose;
			break;
		  case 'v':
			Loglevel = (uint) strtoul(optarg, NULL, 0);
			break;
		  default:
			usage(argv[0]);
			return 1;
		}
	}

	sm_test_begin(argc, argv, "test dnsbl 0");
	argc -= optind;
	argv += optind;
	if (argc > 0) {
		for (i = 0; i < argc; i++)
			t_async_0(argv[i], lflags);
	}
	else
		t_async_0("127.0.0.1", lflags);
	return sm_test_end();
}
#else /* SMAR_USE_DNS && SMAR_ASYNC */
int
main(int argc, char *argv[])
{
	sm_io_fprintf(smioout, "%s: SKIPPED: need SMAR_USE_DNS && SMAR_ASYNC\n",
		argv[0]);
	sm_io_flush(smioout);
	return 0;
}
#endif /* SMAR_USE_DNS && SMAR_ASYNC */
